{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# $i$-RevNet and $i$-ResNet: Minor modifications to get an invertible neural network!"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Our objective will be to design some neural networks that will be invertible."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "import torch.optim as optim\n",
    "import torch.nn.functional as F\n",
    "import torch.nn as nn"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We first consider the model architecture given by: (it's a VGG-like model)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Net(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(Net, self).__init__()\n",
    "        self.conv1 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.conv2 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.conv3 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.conv4 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.fc = nn.Linear(28*28, 10)\n",
    "    def forward(self, x):\n",
    "        x = F.relu(self.conv1(x))\n",
    "        x = F.relu(self.conv2(x))\n",
    "        x = F.relu(self.conv3(x))\n",
    "        x = F.relu(self.conv4(x))\n",
    "        x = x.view(-1, 28*28)\n",
    "        x = self.fc(x)\n",
    "        return x"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The routines to load the data from the dataset are given by:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torchvision\n",
    "import torchvision.transforms as transforms\n",
    "\n",
    "from torch.utils.data import Subset\n",
    "\n",
    "reshape = transforms.Lambda(lambda x:x.view(16,7,7))\n",
    "\n",
    "dataset=torchvision.datasets.MNIST('./', download=True, transform=transforms.Compose([transforms.ToTensor(),reshape]), train=True)\n",
    "train_indices = torch.arange(0,500)#torch.randperm(len(dataset))[:500]\n",
    "train_dataset = Subset(dataset, train_indices)\n",
    "\n",
    "dataset=torchvision.datasets.MNIST('./', download=True, transform=transforms.Compose([transforms.ToTensor(),reshape]), train=False)\n",
    "test_indices = torch.arange(0,10000)#torch.randperm(len(dataset))[:1000]\n",
    "test_dataset = Subset(dataset, test_indices)\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "and the dataloader, which is actually used to depile the data is given by:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "trainloader = torch.utils.data.DataLoader(train_dataset, batch_size=64,\n",
    "                                          shuffle=True, num_workers=0)\n",
    "\n",
    "testloader = torch.utils.data.DataLoader(test_dataset, batch_size=16,\n",
    "                                          shuffle=False, num_workers=0)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "1. Try to visualize some data that we will use along this lab session."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "2. Write a train function and a test function on this dataset. Train the neural network, and plot the training loss. __Hint:__ Use https://pytorch.org/tutorials/beginner/blitz/cifar10_tutorial.html ."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def train(net, n_epoch=40, learning_rate=0.05):\n",
    "    \n",
    "    loss = nn.CrossEntropyLoss()\n",
    "    optimizer = optim.SGD(net.parameters(), lr=learning_rate, momentum=0.9)\n",
    "    \n",
    "    for epoch in range(n_epoch):  # loop over the dataset multiple times\n",
    "        running_loss = []\n",
    "        for i, data in enumerate(trainloader, 0):\n",
    "            # get the inputs; data is a list of [inputs, labels]\n",
    "            inputs, labels = data\n",
    "            \n",
    "            # FILL IN HERE\n",
    "\n",
    "\n",
    "    correct = 0\n",
    "    total = 0\n",
    "    with torch.no_grad():\n",
    "        for data in trainloader:\n",
    "            images, labels = data\n",
    "            outputs = net(images)\n",
    "            _, predicted = torch.max(outputs.data, 1)\n",
    "            total += labels.size(0)\n",
    "            correct += (predicted == labels).sum().item()\n",
    "\n",
    "    print('Accuracy of the network on the 500 train images: %d %%' % (\n",
    "        100 * correct / total))\n",
    "    \n",
    "    return net\n",
    "\n",
    "def test(net):\n",
    "    correct = 0\n",
    "    total = 0\n",
    "    # FILL IN HERE\n",
    "    \n",
    "net = Net()\n",
    "net = train(net)\n",
    "test(net)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# $i$-RevNets: NNs invertible by constructions\n",
    "3. For a layer $j$, assume the features at layer write $X_j=[x_j,\\tilde x_j]$ where $\\text{dim}(x_j)=n_j$ and $\\text{dim}(\\tilde x_j)=m_j$ , let $F_j:\\mathbb{R}^{n_j}\\rightarrow \\mathbb{R}^{m_{j+1}}$. Propose an invertible $\\Phi_F$ architecture using only $F=(F_1,...F_J)$ and additions. We write $\\mathcal{F}_J=\\cup_{F} \\{\\Phi_{F}\\}$ the set of such neural networks of depth $J$ for any width $m_j,n_j$."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "4. Write $\\mathcal{F}=\\cup_J \\mathcal{F}_J$. Give a necessary and suffisant condition for $\\mathcal{F}$ to be a group."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "5. Write a function that splits a tensor of size $(batch, chan, N, N)$ into two tensors of size $(batch, chan//2, N, N)$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SplitAlongChannels(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(SplitAlongChannels, self).__init__()\n",
    "        pass\n",
    "    def forward(self,x):\n",
    "        pass\n",
    "    def inverse(self,X):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "6. Write the corresponding invertible operation."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class InvertibleLayer(nn.Module):\n",
    "    def __init__(self, kernel_size=7):\n",
    "        super(InvertibleLayer, self).__init__()\n",
    "        self.conv = nn.Conv2d(8, 8, kernel_size,padding=kernel_size//2)\n",
    "    def forward(self, X):\n",
    "        pass\n",
    "    def inverse(self, X):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "7. Write a class i-RevNet that consists in 4 layers, which is analog to *Net*."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class iRevNet(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(iRevNet, self).__init__()\n",
    "        self.split = SplitAlongChannels()\n",
    "        self.conv1 = InvertibleLayer()\n",
    "        self.conv2 = InvertibleLayer()\n",
    "        self.conv3 = InvertibleLayer()\n",
    "        self.conv4 = InvertibleLayer()\n",
    "        self.fc = nn.Linear(28*28, 10)\n",
    "        \n",
    "    def forward(self, x, classif=True):\n",
    "        if classif:\n",
    "            pass\n",
    "        pass\n",
    "    \n",
    "    def inverse(self, x):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "8. Implement and run such architecture using the questions above. Compare the accuracies."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "net = iRevNet()\n",
    "net = train(net)\n",
    "test(net)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "9. Validate on an example that your model is exactly invertible."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "x = torch.randn(1,16,7,7)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This type of model is directly linked to: https://openreview.net/pdf?id=HJsjkMb0Z."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# $i$-ResNets: a simple modification to get an invertible resnet"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "We know consider the following class model, to which we refer as ResNet (see https://arxiv.org/abs/1512.03385):"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class ResNet(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(ResNet, self).__init__()\n",
    "        self.conv1 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.conv2 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.conv3 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.conv4 = nn.Conv2d(16, 16, 7,padding=3)\n",
    "        self.fc = nn.Linear(28*28, 10)\n",
    "    def forward(self, x):\n",
    "        x = x - F.relu(self.conv1(x))\n",
    "        x = x - F.relu(self.conv2(x))\n",
    "        x = x - F.relu(self.conv3(x))\n",
    "        x = x - F.relu(self.conv4(x))\n",
    "        x = x.view(-1, 28*28)\n",
    "        x = self.fc(x)\n",
    "        return x"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Note that those models have some identity loop, to which we refer as skip-connections, that were introduced in https://arxiv.org/abs/1512.03385."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "10. Train the corresponding model."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "net=ResNet()\n",
    "net = train(net)\n",
    "test(net)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "11. Let's consider the layer of a Neural Net given by $x_{j+1}=x_j - F_j x_j$. Propose a condition on $F=(F_1,...,F_J)$ for $\\Phi x$ to be invertible. Propose a simple iterative way to compute the inverse. How many iterations are needed to obtain a precision $\\epsilon$?"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "12. Let be $N$ an integer,  $x\\in \\ell^2([0,N-1])$. Compute the norm of $W:y\\rightarrow x\\circledast  y$. Deduce a simple renormalization procedure to constraint $\\Vert DF_j\\Vert\\leq\\rho$ for some $0<\\rho<1$."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "13. Implement this layer and incorporate the normaliztion in the train function. You might want to look at the help of nn.Conv2D to access the weight of the conv. __Hint__: modify a parameter using .data to avoid breaking the tree of computation of pytorch."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from torch.fft import fft\n",
    "from torch.fft import ifft\n",
    "class InvertibleResidual(nn.Module):\n",
    "    \n",
    "    def __init__(self, kernel_size=7):\n",
    "        super(InvertibleResidual, self).__init__()\n",
    "        self.conv = nn.Conv2d(16, 16, kernel_size,padding=kernel_size//2)\n",
    "        \n",
    "    def forward(self, x):\n",
    "        pass\n",
    "    \n",
    "    def normalize(self, rho=0.9):\n",
    "        pass\n",
    "    \n",
    "    def inverse(self,y):\n",
    "        pass"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def train_norm(net, n_epoch=40, learning_rate=0.05):\n",
    "    \n",
    "   # Fill in\n",
    "    \n",
    "    return net"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "14. Write a class i-ResNet, and train it."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class iResNet(nn.Module):\n",
    "    def __init__(self):\n",
    "        super(iResNet, self).__init__()\n",
    "        self.conv1 = InvertibleResidual()\n",
    "        self.conv2 = InvertibleResidual()\n",
    "        self.conv3 = InvertibleResidual()\n",
    "        self.conv4 = InvertibleResidual()\n",
    "        self.fc = nn.Linear(28*28, 10)\n",
    "    \n",
    "    def normalize(self, rho=0.9):\n",
    "        # FILL IN\n",
    "        return self\n",
    "\n",
    "    def forward(self, x, classif=True):\n",
    "        pass\n",
    "        if classif:\n",
    "            pass\n",
    "\n",
    "    \n",
    "    def inverse(self, x):\n",
    "        pass\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "net = iResNet()\n",
    "\n",
    "net = train_norm(net)\n",
    "test(net)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "15. Verify the invertibility of the final model."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "x = torch.randn(1,16,7,7)\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "This type of model is quite discussed in https://arxiv.org/pdf/1811.00995.pdf."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "16. For some invertible neural networks $\\Phi$ and two images $x_0,x_1$, visualize for $0<t<1$ $x_t=\\Phi^{-1}((1-t)\\Phi x_0+t\\Phi x_1$. What do you think?"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "17. Explain the major differences between the two approaches, the pros and the cons."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.16"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
